from moai.supervision.weighted import Weighted
from moai.supervision.noop import NoOp
from moai.supervision.homoscedastic import Homoscedastic

__all__ = [
    "Weighted",
    "NoOp",
    'Homoscedastic',
]